unit TedNetUnit;

// Neural MLP network
//
// (c)2003, TEDSoft GmbH, Bonn, Germany
//
// www.tedsoft.de, info@tedsoft.de
//
// Alle Rechte vorbehalten.
// 
// All rights reserved - unless granted by written agreement.
//
// It is not permitted to remove this notice.
//
// Free for noncommercial use.

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls;

type
  TDoubleArray = array of double;

  //data structures for training the net

  TTrainingValue   = record
                       input: TDoubleArray;
                       output: TDoubleArray;
                     end;

  TTedNet = class;

  TTrainingPattern = class(TPersistent)
    private
      FNet: TTedNet;
      FTrainingValues: array of TTrainingValue;
      function GetTrainingValue(index: integer): TTrainingValue;
      procedure SetTrainingValue(index: integer;
                const trainingValueIn: TTrainingValue);

      //loading and saving to .dfm file
      procedure LoadTrainingValues (Reader: TReader);
      procedure SaveTrainingValues (Writer: TWriter);
    protected
      procedure DefineProperties (Filer: TFiler); override;
    public
      constructor create();
      destructor destroy(); override;

      procedure Clear;

      //copy routine
      procedure assign(source: TPersistent); override;

      function getNumberOfTrainingValues(): integer;

      //import complete training pattern
      procedure SetPatternWithDoubles(const doubles: TDoubleArray; var
                        numberOfTrainingValues: integer;
                        var trainingInit: boolean);
      property TrainingValues[index: integer]: TTrainingValue read
                  GetTrainingValue write SetTrainingValue; default;
  end;
  //end of TTrainingPattern declaration

  //beginning of declaration for topology
  //topology of the network: (#input, #hidden, #output)
  TTopology = class(TPersistent)
  protected
    FNumberOfInputNeurons: Byte;
    FNumberOfHiddenNeurons: Byte;
    FNumberOfOutputNeurons: Byte;
    FTedNet: TTedNet;

    procedure SetNumberOfInputNeurons(const NumbOfInputsIn: Byte); virtual;
    procedure SetNumberOfHiddenNeurons(const NumbOfHidIn: Byte); virtual;
    procedure SetNumberOfOutputNeurons(const NumbOfOutputsIn: Byte); virtual;

  public
    constructor Create(); virtual;
    destructor Destroy(); override;

    //initialising FTedNet
    procedure SetTedNet(const TedNetIn: TTedNet);

  published
    property NumberOfInputNeurons: Byte read FNumberOfInputNeurons write
                                                SetNumberOfInputNeurons;
    property NumberOfHiddenNeurons: Byte read FNumberOfHiddenNeurons write
                                                SetNumberOfHiddenNeurons;
    property NumberOfOutputNeurons: Byte read FNumberOfOutputNeurons write
                                                SetNumberOfOutputNeurons;
  end;
  //end of TTopology declaration

  //beginning of TExtendedTopology declaration
  TExtendedTopology = class(TTopology)
  private
    FNumberOfHidden2Neurons: Byte; //for second hidden layer
    FHiddenLayer2Active: Boolean;  //second hidden layer is optional

    //for changing the properties
    procedure SetNumberOfHidden2Neurons(const NumbOfHid2In: Byte);
    procedure SetSecondHiddenLayer(const addLayer: Boolean);

  public
    constructor Create(); override;
    destructor Destroy(); override;

  published
    property NumberOfInputNeurons: Byte read FNumberOfInputNeurons write
                                                SetNumberOfInputNeurons;
    property NumberOfHiddenNeurons: Byte read FNumberOfHiddenNeurons write
                                                SetNumberOfHiddenNeurons;
    property NumberOfHidden2Neurons: Byte read FNumberOfHidden2Neurons write
                                                SetNumberOfHidden2Neurons;
    property NumberOfOutputNeurons: Byte read FNumberOfOutputNeurons write
                                                SetNumberOfOutputNeurons;
    property HiddenLayer2Active: Boolean read FHiddenLayer2Active write
                                                SetSecondHiddenLayer;

  end;
  //end of TExtendedTopology declaration

  TTedNet = class(TComponent)
    protected
      FTrainingInit: Boolean;
      FNetInit: Boolean;
      FTrainingPattern: TTrainingPattern;
    public
      constructor create(AOwner: TComponent); override;
      destructor destroy(); override;
      //returns associated topology
      function getTopologyAsTTopology: TTopology; virtual; abstract;
  end;

  //parses text file with lines of doubles, separated by blancs
  //comment lines start with #
  //numberOfLines gets actual number of lines containing doubles
  procedure parseTextFileIntoDoubles(const fileName: string;
                        var doubles: TDoubleArray; var numberOfLines: integer);
  function distanceSquare(x,c: TDoubleArray): double;  //euclidic distance
                                                      //squared
  procedure errorHandling(errorIndex: byte);
implementation

//implementation of TTopology
constructor TTopology.Create();
begin
  inherited Create();

  //defaults
  FNumberOfInputNeurons:=2;
  FNumberOfHiddenNeurons:=3;
  FNumberOfOutputNeurons:=4;
  FTedNet := nil;
end;

procedure TTopology.SetNumberOfInputNeurons(const NumbOfInputsIn: Byte);
begin
  if ((NumbOfInputsIn > 0) and (NumbOfInputsIn <> FNumberOfInputNeurons)) then
  begin
    FNumberOfInputNeurons:=NumbOfInputsIn;

    // new trainingpattern, new init of TedCustomNet required
    if assigned(FTedNet) then
    begin
      FTedNet.FTrainingInit := False;
      FTedNet.FNetInit := False;
    end;
  end;
end;

procedure TTopology.SetNumberOfHiddenNeurons(const NumbOfHidIn: Byte);
begin
  if ((NumbOfHidIn > 0) and (NumbOfHidIn <> FNumberOfHiddenNeurons)) then
  begin
    FNumberOfHiddenNeurons:=NumbOfHidIn;

    // new init of TedRBFnet required
    if assigned(FTedNet) then
    begin
      FTedNet.FNetInit := False;
    end;
  end;
end;

procedure TTopology.SetNumberOfOutputNeurons(const NumbOfOutputsIn: Byte);
begin
  if ((NumbOfOutputsIn > 0) and (NumbOfOutputsIn <> FNumberOfOutputNeurons))then
  begin
    FNumberOfOutputNeurons:=NumbOfOutputsIn;

    // new trainingpattern, new init of TedRBFnet required
    if assigned(FTedNet) then
    begin
      FTedNet.FTrainingInit := False;
      FTedNet.FNetInit := False;
    end;
  end;
end;

destructor TTopology.Destroy();
begin
  inherited Destroy();
end;
//end TTopology implementation

//implementation of TExtendedTopology
constructor TExtendedTopology.Create();
begin
  inherited create;

  //if no second hidden layer, number of neurons in layer 2 = 0
  FNumberOfHidden2Neurons:=0;
  FHiddenLayer2Active:=FALSE;
end;

procedure TExtendedTopology.SetSecondHiddenLayer(const addLayer: Boolean);
begin
  if (FHiddenLayer2Active <> addLayer) then
  begin
    FHiddenLayer2Active:=addLayer;
    FNumberOfHidden2Neurons:=0;

    if assigned(FTedNet) then
    begin
      FTedNet.FNetInit:=False;
    end;
  end;
end;

procedure TExtendedTopology.SetNumberOfHidden2Neurons(const NumbOfHid2In: Byte);
begin
  if FHiddenLayer2Active and (NumbOfHid2In > 0)
    and (NumbOfHid2In<>FNumberOfHidden2Neurons) then
    FNumberOfHidden2Neurons:=NumbOfHid2In;

  if assigned(FTedNet) then
  begin
    FTedNet.FNetInit:=False;
  end;
end;

destructor TExtendedTopology.Destroy();
begin
  inherited Destroy();
end;
//end TExtendedTopology definition

function TTrainingPattern.GetTrainingValue(index: integer): TTrainingValue;
begin
  if (Index < 0) OR (index >= length(FTrainingValues)) then
    raise Exception.Create ('Index out of bounds.')
  else begin
    GetTrainingValue.input:=copy(FTrainingValues[index].input);
    GetTrainingValue.output:=copy(FTrainingValues[index].output);
  end
end;

constructor TTrainingPattern.create();
begin
  inherited create;
  FNet:=nil;
  FTrainingValues:=nil;
end;

destructor TTrainingPattern.destroy();
begin
  inherited destroy;
end;

procedure TTrainingPattern.SetTrainingValue(index: integer;
            const trainingValueIn: TTrainingValue);
begin
  if (Index < 0) OR (Index > Length (FTrainingValues)) then
    raise Exception.Create ('Index out of bounds.');
  if Index = Length (FTrainingValues) then
    SetLength (FTrainingValues, Length (FTrainingValues) + 1);
  ftrainingValues[index].input:=copy(trainingValueIn.input);
  ftrainingValues[index].output:=copy(trainingValueIn.output);
end;

function TTrainingPattern.getNumberOfTrainingValues(): integer;
begin
  Result:=length(ftrainingValues);
end;

procedure TTrainingPattern.assign(source: TPersistent);
var
  i: integer;
  tmp: TTrainingPattern;
begin
  if source = nil then
  begin
    clear;
    FNet:=nil;
  end
  else
    if not (source is TTrainingPattern) then
      inherited Assign (source)
    else
    begin
      tmp := source as TTrainingPattern;
      FNet:=tmp.FNet;
      setLength(fTrainingValues,Length(tmp.fTrainingValues));
      for i:=0 to Length (tmp.FTrainingValues)-1 do
      begin
        fTrainingValues[i].input:=copy(tmp.FTrainingValues[i].input);
        fTrainingValues[i].output:=copy(tmp.FTrainingValues[i].output);
    end;
  end;
end;


procedure TTrainingPattern.SetPatternWithDoubles
        (const doubles: TDoubleArray; var numberOfTrainingValues: integer;
         var trainingInit: boolean);
var
  counterOfDoubles, i,j: integer;
  topology: TTopology;
begin
  //check for undefined data
  if (FNet=nil) then
  begin
    errorHandling(1);
    exit;
  end;
  topology:=FNet.getTopologyAsTTopology;
  if (length(doubles) mod (topology.NumberOfInputNeurons+
   Topology.NumberOfOutputNeurons)> 0) or
   (length(doubles) div (Topology.NumberOfInputNeurons+
   Topology.NumberOfOutputNeurons)<>numberOfTrainingValues)
   then
  begin
    errorHandling(2);
    trainingInit:=false;
    exit;
  end;

  //get memory for FTrainingpattern
  setLength(FTrainingValues,numberOfTrainingValues);
  for i:=0 to numberOfTrainingValues-1 do
  begin
    setLength(FTrainingValues[i].input,
          Topology.numberOfInputNeurons);
    setLength(FTrainingValues[i].output,
          Topology.numberOfOutputNeurons);
  end;

  counterOfDoubles:=0;
  for i:=0 to numberOfTrainingValues-1 do
  begin
    for j:=0 to Topology.NumberOfInputNeurons-1 do
    begin
      FTrainingValues[i].input[j]:=doubles[counterOfdoubles];
      counterOfDoubles:=counterOfDoubles+1;
    end;
    for j:=0 to Topology.NumberOfOutputNeurons-1 do
    begin
      FTrainingValues[i].output[j]:=doubles[counterOfdoubles];
      counterOfDoubles:=counterOfDoubles+1;
    end;
  end;
  FNet.FTrainingInit:=true;
  TrainingInit:=true;
end;

procedure TTrainingPattern.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineProperty('TrainingValues', LoadTrainingValues, SaveTrainingValues,
  True);
end;

procedure TTrainingPattern.LoadTrainingValues(Reader: TReader);
var
  i, j: Integer;
  ii, jj: Integer;
  trainingValue: TTrainingValue;
begin
  Clear;  //erasing current trainingPattern
  Reader.ReadListBegin;
  i := Reader.ReadInteger;
  for ii := 0 to i - 1 do
  begin
    j := Reader.ReadInteger;
    SetLength (trainingValue.input, j);
    Reader.ReadListBegin;
    for jj := 0 to j - 1 do
      trainingValue.input[jj] := Reader.ReadFloat;
    Reader.ReadListEnd;
    j := Reader.ReadInteger;
    SetLength (trainingValue.output, j);
    Reader.ReadListBegin;
    for jj := 0 to j - 1 do
      trainingValue.output[jj] := Reader.ReadFloat;
    SetTrainingValue(ii, trainingValue);
    Reader.ReadListEnd;
  end;
  Reader.ReadListEnd;
end;

procedure TTrainingPattern.SaveTrainingValues(Writer: TWriter);
var
  i, j: Integer;
  trainingValue: TTrainingValue;
begin
  Writer.WriteListBegin;
  Writer.WriteInteger(getNumberOfTrainingValues);
  for i := 0 to getNumberOfTrainingValues - 1 do
  begin
    trainingValue := GetTrainingValue (i);
    Writer.WriteInteger (Length (trainingValue.input));
    Writer.WriteListBegin;
    for j := 0 to Length (trainingValue.input) - 1 do
    begin
      Writer.WriteFloat (trainingValue.input[j]);
    end;
    Writer.WriteListEnd;
    Writer.WriteInteger (Length (trainingValue.output));
    Writer.WriteListBegin;
    for j := 0 to Length (trainingValue.output) - 1 do
    begin
      Writer.WriteFloat (trainingValue.output[j]);
    end;
    Writer.WriteListEnd;
  end;
  Writer.WriteListEnd;
end;

procedure TTrainingPattern.Clear;
begin
  FTrainingValues := nil;
end;
//end of TTrainingPattern definition


constructor TTedNet.create(AOwner: TComponent);
begin
  inherited create(AOwner);

  FTrainingPattern := TTrainingPattern.Create();
  FTrainingPattern.FNet := Self;
end;

destructor TTedNet.destroy();
begin
  inherited destroy;
end;

procedure TTopology.SetTedNet(const TedNetIn: TTedNet);
begin
  FTedNet:=TedNetIn;
end;

procedure parseTextFileIntoDoubles(const fileName: string;
                        var doubles: TDoubleArray;var numberOfLines: integer);
var
  posInLine, counterOfNumbers, counterOfLines: integer;
  f: textfile;
  line, numberInString: string;
begin
  DecimalSeparator:='.';
  assignFile(f,fileName);
  reset(f);
  counterOfNumbers:=0;
  counterOfLines:=0;
  setLength(numberInString,20); //numbers must not be longer than 20 characters
  while (not eof(f)) do
  begin
    readln(f,line);
    //its not a comment, isn't it?
    if not (line[1]='#') then
    begin
      counterOfLines:=counterOfLines+1;
      numberInString:='';
      for posInLine:=1 to length(line) do
      begin
        if  (line[posInLine]=' ') then
        begin
          // is there a number to convert?
          if not (numberInString='') then
          begin
            counterOfNumbers:=counterOfNumbers+1;
            //get memory for new double
            setLength(doubles, counterOfNumbers);
            doubles[counterOfNumbers-1]:=StrToFloat(numberInString);
            numberInString:='';
          end;
        end
        else
        begin
          //scan the number
          numberInString:=numberInString+line[posInLine];
        end;
      end;
      // last number is probably not yet converted
      if not (numberInString='') then
      begin
        counterOfNumbers:=counterOfNumbers+1;
        setLength(doubles, counterOfNumbers);
        doubles[counterOfNumbers-1]:=StrToFloat(numberInString);
        numberInString:='';
      end;
    end;
  end;
  closeFile(f);
  numberOfLines:=counterOfLines;
end;

function distanceSquare(x,c: TDoubleArray): double;
var
   i: Integer;
   d: double;
begin
     d:=0;

     for i:=0 to Length(x)-1 do
         d := d + (x[i]-c[i]) * (x[i]-c[i]);

     DistanceSquare := d;
end;

procedure errorHandling(errorIndex: byte);
begin
  case errorIndex of
    1: showMessage('No net available');
    2: showMessage('Invalid training data');
  end;
end;

end.
